/*
io_sd_card.c

Hardware routines for interfacing with SD cards on a Ninjapass X9

Copyright (C) 2007 by CJ Bell
siegebell@gmail.com


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
Boston, MA  02110-1301, USA.
*/

#include <nds.h>
//#include <stdio.h>
#include "io_x9_card.h"
#include "io_sd_card.h"



void SDSendCommand(SDCommand sdCmd, uint32 arg, uint8* response)
{
    // Each byte in var represent one bit -- bit 0 of var[x]. Thus, each group
    // of eight bytes in var is "effectively" one byte from the SD card
    // 
    uint8 var[0x200];
    X9CardIO(0x60, sdCmd, arg, 0xA1586000, (uint32*)var, 0x80);

    // Skips the first 7 effective bytes (k = 0x30)
    // 0x30 is the size in bits of an SD-mode response
    // the rest must be resulting data...
    int effectiveBit = 0x30;

    // Search for the first "1"/high bit
    // (This is used to find the first byte, which must start with a "1")
    int start;
    int end;
    while(true)
    {
        if(!(var[effectiveBit] & 1))
        {// start-bit
            // Start at the byte found
            start = effectiveBit;
            // Use 256 bytes of var to construct 32 effective bytes
            end = effectiveBit + 0x100;
            break;
        }

        ++effectiveBit;

        if(effectiveBit >= 0x200)
        {// Hit the end of the search without success 
            // Skip the start-bit
            start = 1; //(var[effectiveBit-1] ^ 1) & 1; <-- known because of previous if-statement
            // Use 255 bytes of var to construct 31+7/8 effective bytes
            end = 0x100; // effectiveBit - 0x100 --> 0x200 - 0x100 --> 0x100
            break;
        }
    }

    // Covert 256 bytes of var into 32 bytes using the low-bits
    // This doesn't care what sd_cs_data is initially because each entry will
    // be shifted 8 bits.
    if(start < end)
    {
        const uint8* end_ptr = &(var[end]);
        uint8* ptr = &(var[start]);

        int bit = 7;
        int j = 0;
        while(ptr < end_ptr)
        {
            if(bit < 0)
            {// Go to the next effective byte
                bit = 7;
                ++j;
            }

            response[j] = (response[j] << 1) | (*ptr & 1);

            --bit;
            ++ptr;
        } 

    }

}

bool SDWriteSingleBlock(uint32 address, const void* buffer)
//bool SDWriteSingleBlock(uint32 address, u32 * buffer)
{
	u8 temp[512];

    // Preemptively send the data before issuing the command so
    // that the X9 can compute the CRC
    X9CardWriteData(0, 0, buffer);

    uint8 response[0x20];

    SDSendCommand(WriteSingleBlock,address,response);

    X9CardIO(0x63, 0, 0, 0xB15863FA, (u32 *)temp, 0x80);

    // TODO: Timeout
    int idx;
    do
    {
        X9CardIO(0x62, 0, 0, 0xA1586000, (u32 *)temp, 0x80);

        for(idx = 0; idx < 0x200 && (temp[idx] & 0x10); ++idx)
            ;

    } while(idx == 0x200);


    // TODO: TIMEOUT
    do
    {
       X9CardIO(0x62, 0, 0, 0xA1586000, (u32 *)temp, 0x80);

        for(idx = 0; idx < 0x200 && !(temp[idx] & 0x10); ++idx)
            ;

    } while(idx == 0x200);

    return true;

}


bool SDReadSingleBlock(uint32 address, void* destination)
{
    uint8 buffer[0x1000];// split in two halves

    X9CardIO(0x60, (uint8)ReadSingleBlock, address, 0xA1586000, (uint32*)buffer, 0x80);

    unsigned int k = 0;
    for(k = 0; k < 0x200 && (buffer[k] & 0xf0); ++k)
        ;

    unsigned int  c = 0;
    while(k == 0x200)
    {
        X9CardIO(0x62, 0, address, 0xA1586000, (uint32*)buffer, 0x80);
        
        c+= 0x80;

        for(k = 0; k < 0x200 && (buffer[k] & 0xf0); ++k)
            ;
    }

    c = k + 1; // just after the 0

    X9CardIO(0x62, 0, address, 0xA2586000, (uint32*)&buffer[0x200], 0x100);

    k+= 0x400; // read 0x400 half-bytes, or 0x200 bytes
    ++k;

    // IO is done with half-bytes: reconstruct full bytes
    uint8* dest = (uint8*)destination;
    for(; c < k; c+= 2)
        *dest++ = (buffer[c] & 0xF0) | (buffer[c+1] >> 4);
    X9CardIO(0x62, 0, address, 0xA1586000, (uint32*)buffer, 0x80);

    return true;
}

bool SDInitialize(unsigned int timeout, uint32* relativeCardAddress)
{
    if(!relativeCardAddress)
        return false;

    int attempts;
    uint8 response[0x20];

    // CMD0 initialize (reset)
    SDSendCommand(SetIdle,0,response);

    // TODO: this should return the correct voltage: use that
    SDSendCommand(AppCommand,0,response);
    SDSendAppCommand(ACMD41,0x00200000,response); // 3.3-3.4 VDD Voltage Window


    //ACMD41 - identify, wait til ready..
    // Should be repeated N times until success, else fail
    // set voltage?
    // sd_cs_data set by sendSDCommand()
    for(attempts = 0; response[1] != 0x80 /* Ready */; ++attempts)
    {
        if(attempts >= timeout)
            return false;

        SDSendCommand(AppCommand,0,response);
        SDSendAppCommand(ACMD41,0x00100000,response); // 3.2-3.3 VDD Voltage Window
    }

    // Retrieves Card's Name
    // CMD2 get CID
    SDSendCommand(GetCID,0,response);

    // Should be repeated N time for success, else fail
    attempts = 0;
    do
    {
        SDSendCommand(GetRCA,0,response);
        ++attempts;
    } while(attempts < timeout && (response[3] & 0x1e) == (SDStandbyState << 1)); // busy

    if(attempts >= timeout)
        return false;

    *relativeCardAddress = (response[1] << 24) | (response[2] << 16);



    // Some cards won't go to higher speeds unless they think you
    // checked their capabilities
    SDSendCommand(GetCSD,*relativeCardAddress,response);

    //AL go SD data mode
    SDSendCommand(SelectCard,*relativeCardAddress,response);


    //AL set SD bus width
    SDSendCommand(AppCommand,*relativeCardAddress,response);
    SDSendAppCommand(SetBusWidth,2,response); // 4-bit bus

    // Use 512 byte blocks
    SDSendCommand(SetBlockLength,512,response);

    return true;
}

